@@ -27,7 +27,7 @@
    utility has right now is to be able to skip them gracefully and allow the
    compilation process to continue.
 
-   That said, see examples/clang_asm_normalize/ for a solution that may
+   That said, see utils/clang_asm_normalize/ for a solution that may
    allow clang users to make things work even with hand-crafted assembly. Just
    note that there is no equivalent for GCC.
 
@@ -47,16 +47,17 @@
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
+#include <limits.h>
 #include <ctype.h>
 #include <fcntl.h>
 
 #include <sys/wait.h>
 #include <sys/time.h>
 
-static u8** as_params;              /* Parameters passed to the real 'as'   */
+static u8 **as_params;              /* Parameters passed to the real 'as'   */
 
-static u8* input_file;              /* Originally specified input file      */
-static u8* modified_file;           /* Instrumented file for the real 'as'  */
+static u8 *input_file;              /* Originally specified input file      */
+static u8 *modified_file;           /* Instrumented file for the real 'as'  */
 
 static u8 be_quiet,                 /* Quiet mode (no stderr output)        */
     clang_mode,                     /* Running in clang mode?               */
@@ -79,9 +80,9 @@ static u8 use_64bit = 1;
 
 static u8 use_64bit = 0;
 
-#ifdef __APPLE__
-#error "Sorry, 32-bit Apple platforms are not supported."
-#endif                                                         /* __APPLE__ */
+  #ifdef __APPLE__
+    #error "Sorry, 32-bit Apple platforms are not supported."
+  #endif                                                       /* __APPLE__ */
 
 #endif                                                     /* ^WORD_SIZE_64 */
 
@@ -89,7 +90,7 @@ static u8 use_64bit = 0;
    is always the last parameter passed by GCC, so we exploit this property
    to keep the code simple. */
 
-static void edit_params(int argc, char** argv) {
+static void edit_params(int argc, char **argv) {
 
   u8 *tmp_dir = getenv("TMPDIR"), *afl_as = getenv("AFL_AS");
   u32 i;
@@ -126,28 +127,38 @@ static void edit_params(int argc, char** argv) {
      is not set. We need to check these non-standard variables to properly
      handle the pass_thru logic later on. */
 
-  if (!tmp_dir) tmp_dir = getenv("TEMP");
-  if (!tmp_dir) tmp_dir = getenv("TMP");
-  if (!tmp_dir) tmp_dir = "/tmp";
+  if (!tmp_dir) { tmp_dir = getenv("TEMP"); }
+  if (!tmp_dir) { tmp_dir = getenv("TMP"); }
+  if (!tmp_dir) { tmp_dir = "/tmp"; }
 
-  as_params = ck_alloc((argc + 32) * sizeof(u8*));
+  as_params = ck_alloc((argc + 32) * sizeof(u8 *));
+  if (unlikely((INT_MAX - 32) < argc || !as_params)) {
 
-  as_params[0] = afl_as ? afl_as : (u8*)"as";
+    FATAL("Too many parameters passed to as");
+
+  }
+
+  as_params[0] = afl_as ? afl_as : (u8 *)"as";
 
   as_params[argc] = 0;
 
-  for (i = 1; i < argc - 1; i++) {
+  for (i = 1; (s32)i < argc - 1; i++) {
+
+    if (!strcmp(argv[i], "--64")) {
 
-    if (!strcmp(argv[i], "--64"))
       use_64bit = 1;
-    else if (!strcmp(argv[i], "--32"))
+
+    } else if (!strcmp(argv[i], "--32")) {
+
       use_64bit = 0;
 
+    }
+
 #ifdef __APPLE__
 
     /* The Apple case is a bit different... */
 
-    if (!strcmp(argv[i], "-arch") && i + 1 < argc) {
+    if (!strcmp(argv[i], "-arch") && i + 1 < (u32)argc) {
 
       if (!strcmp(argv[i + 1], "x86_64"))
         use_64bit = 1;
@@ -195,11 +206,16 @@ static void edit_params(int argc, char** argv) {
 
     }
 
-    if (input_file[1])
+    if (input_file[1]) {
+
       FATAL("Incorrect use (not called through afl-gcc?)");
-    else
+
+    } else {
+
       input_file = NULL;
 
+    }
+
   } else {
 
     /* Check if this looks like a standard invocation as a part of an attempt
@@ -210,15 +226,20 @@ static void edit_params(int argc, char** argv) {
     if (strncmp(input_file, tmp_dir, strlen(tmp_dir)) &&
         strncmp(input_file, "/var/tmp/", 9) &&
         strncmp(input_file, "/tmp/", 5) &&
-        getenv("AFL_AS_FORCE_INSTRUMENT") == NULL)
+        getenv("AFL_AS_FORCE_INSTRUMENT") == NULL) {
+
       pass_thru = 1;
-    else if (getenv("AFL_AS_FORCE_INSTRUMENT"))
+
+    } else if (getenv("AFL_AS_FORCE_INSTRUMENT")) {
+
       unsetenv("AFL_AS_FORCE_INSTRUMENT");
 
+    }
+
   }
 
-  modified_file =
-      alloc_printf("%s/.afl-%u-%u.s", tmp_dir, getpid(), (u32)time(NULL));
+  modified_file = alloc_printf("%s/.afl-%u-%u-%u.s", tmp_dir, (u32)getpid(),
+                               (u32)time(NULL), (u32)random());
 
 wrap_things_up:
 
@@ -234,8 +255,8 @@ static void add_instrumentation(void) {
 
   static u8 line[MAX_LINE];
 
-  FILE* inf;
-  FILE* outf;
+  FILE *inf;
+  FILE *outf;
   s32   outfd;
   u32   ins_lines = 0;
 
@@ -244,26 +265,28 @@ static void add_instrumentation(void) {
 
 #ifdef __APPLE__
 
-  u8* colon_pos;
+  u8 *colon_pos;
 
 #endif                                                         /* __APPLE__ */
 
   if (input_file) {
 
     inf = fopen(input_file, "r");
-    if (!inf) PFATAL("Unable to read '%s'", input_file);
+    if (!inf) { PFATAL("Unable to read '%s'", input_file); }
 
-  } else
+  } else {
 
     inf = stdin;
 
-  outfd = open(modified_file, O_WRONLY | O_EXCL | O_CREAT, 0600);
+  }
 
-  if (outfd < 0) PFATAL("Unable to write to '%s'", modified_file);
+  outfd = open(modified_file, O_WRONLY | O_EXCL | O_CREAT, DEFAULT_PERMISSION);
+
+  if (outfd < 0) { PFATAL("Unable to write to '%s'", modified_file); }
 
   outf = fdopen(outfd, "w");
 
-  if (!outf) PFATAL("fdopen() failed");
+  if (!outf) { PFATAL("fdopen() failed"); }
 
   while (fgets(line, MAX_LINE, inf)) {
 
@@ -287,7 +310,7 @@ static void add_instrumentation(void) {
 
     fputs(line, outf);
 
-    if (pass_thru) continue;
+    if (pass_thru) { continue; }
 
     /* All right, this is where the actual fun begins. For one, we only want to
        instrument the .text section. So, let's keep track of that in processed
@@ -300,9 +323,12 @@ static void add_instrumentation(void) {
          around them, so we use that as a signal. */
 
       if (!clang_mode && instr_ok && !strncmp(line + 2, "p2align ", 8) &&
-          isdigit(line[10]) && line[11] == '\n')
+          isdigit(line[10]) && line[11] == '\n') {
+
         skip_next_label = 1;
 
+      }
+
       if (!strncmp(line + 2, "text\n", 5) ||
           !strncmp(line + 2, "section\t.text", 13) ||
           !strncmp(line + 2, "section\t__TEXT,__text", 21) ||
@@ -330,23 +356,23 @@ static void add_instrumentation(void) {
 
     if (strstr(line, ".code")) {
 
-      if (strstr(line, ".code32")) skip_csect = use_64bit;
-      if (strstr(line, ".code64")) skip_csect = !use_64bit;
+      if (strstr(line, ".code32")) { skip_csect = use_64bit; }
+      if (strstr(line, ".code64")) { skip_csect = !use_64bit; }
 
     }
 
     /* Detect syntax changes, as could happen with hand-written assembly.
        Skip Intel blocks, resume instrumentation when back to AT&T. */
 
-    if (strstr(line, ".intel_syntax")) skip_intel = 1;
-    if (strstr(line, ".att_syntax")) skip_intel = 0;
+    if (strstr(line, ".intel_syntax")) { skip_intel = 1; }
+    if (strstr(line, ".att_syntax")) { skip_intel = 0; }
 
     /* Detect and skip ad-hoc __asm__ blocks, likewise skipping them. */
 
     if (line[0] == '#' || line[1] == '#') {
 
-      if (strstr(line, "#APP")) skip_app = 1;
-      if (strstr(line, "#NO_APP")) skip_app = 0;
+      if (strstr(line, "#APP")) { skip_app = 1; }
+      if (strstr(line, "#NO_APP")) { skip_app = 0; }
 
     }
 
@@ -375,16 +401,19 @@ static void add_instrumentation(void) {
      */
 
     if (skip_intel || skip_app || skip_csect || !instr_ok || line[0] == '#' ||
-        line[0] == ' ')
+        line[0] == ' ') {
+
       continue;
 
+    }
+
     /* Conditional branch instruction (jnz, etc). We append the instrumentation
        right after the branch (to instrument the not-taken path) and at the
        branch destination label (handled later on). */
 
     if (line[0] == '\t') {
 
-      if (line[1] == 'j' && line[2] != 'm' && R(100) < inst_ratio) {
+      if (line[1] == 'j' && line[2] != 'm' && R(100) < (long)inst_ratio) {
 
         fprintf(outf, use_64bit ? trampoline_fmt_64 : trampoline_fmt_32,
                 R(MAP_SIZE));
@@ -426,7 +455,7 @@ static void add_instrumentation(void) {
         /* Apple: L<num> / LBB<num> */
 
         if ((isdigit(line[1]) || (clang_mode && !strncmp(line, "LBB", 3))) &&
-            R(100) < inst_ratio) {
+            R(100) < (long)inst_ratio) {
 
 #else
 
@@ -434,7 +463,7 @@ static void add_instrumentation(void) {
 
         if ((isdigit(line[2]) ||
              (clang_mode && !strncmp(line + 1, "LBB", 3))) &&
-            R(100) < inst_ratio) {
+            R(100) < (long)inst_ratio) {
 
 #endif                                                         /* __APPLE__ */
 
@@ -449,11 +478,16 @@ static void add_instrumentation(void) {
              .Lfunc_begin0-style exception handling calculations (a problem on
              MacOS X). */
 
-          if (!skip_next_label)
+          if (!skip_next_label) {
+
             instrument_next = 1;
-          else
+
+          } else {
+
             skip_next_label = 0;
 
+          }
+
         }
 
       } else {
@@ -468,24 +502,27 @@ static void add_instrumentation(void) {
 
   }
 
-  if (ins_lines) fputs(use_64bit ? main_payload_64 : main_payload_32, outf);
+  if (ins_lines) { fputs(use_64bit ? main_payload_64 : main_payload_32, outf); }
 
-  if (input_file) fclose(inf);
+  if (input_file) { fclose(inf); }
   fclose(outf);
 
   if (!be_quiet) {
 
-    if (!ins_lines)
+    if (!ins_lines) {
+
       WARNF("No instrumentation targets found%s.",
             pass_thru ? " (pass-thru mode)" : "");
-    else {
+
+    } else {
 
       char modeline[100];
-      snprintf(modeline, sizeof(modeline), "%s%s%s%s",
+      snprintf(modeline, sizeof(modeline), "%s%s%s%s%s",
                getenv("AFL_HARDEN") ? "hardened" : "non-hardened",
                getenv("AFL_USE_ASAN") ? ", ASAN" : "",
                getenv("AFL_USE_MSAN") ? ", MSAN" : "",
-               getenv("AFL_USE_UBSAN") ? ", UBSAN" : "");
+               getenv("AFL_USE_UBSAN") ? ", UBSAN" : "",
+               getenv("AFL_USE_LSAN") ? ", LSAN" : "");
 
       OKF("Instrumented %u locations (%s-bit, %s mode, ratio %u%%).", ins_lines,
           use_64bit ? "64" : "32", modeline, inst_ratio);
@@ -498,12 +535,12 @@ static void add_instrumentation(void) {
 
 /* Main entry point */
 
-int main(int argc, char** argv) {
+int main(int argc, char **argv) {
 
   s32 pid;
-  u32 rand_seed;
+  u32 rand_seed, i, j;
   int status;
-  u8* inst_ratio_str = getenv("AFL_INST_RATIO");
+  u8 *inst_ratio_str = getenv("AFL_INST_RATIO");
 
   struct timeval  tv;
   struct timezone tz;
@@ -514,10 +551,12 @@ int main(int argc, char** argv) {
 
     SAYF(cCYA "afl-as" VERSION cRST " by Michal Zalewski\n");
 
-  } else
+  } else {
 
     be_quiet = 1;
 
+  }
+
   if (argc < 2 || (argc == 2 && strcmp(argv[1], "-h") == 0)) {
 
     fprintf(
@@ -547,7 +586,7 @@ int main(int argc, char** argv) {
         "AFL_QUIET: suppress verbose output\n"
         "AFL_KEEP_ASSEMBLY: leave instrumented assembly files\n"
         "AFL_AS_FORCE_INSTRUMENT: force instrumentation for asm sources\n"
-        "AFL_HARDEN, AFL_USE_ASAN, AFL_USE_MSAN, AFL_USE_UBSAN:\n"
+        "AFL_HARDEN, AFL_USE_ASAN, AFL_USE_MSAN, AFL_USE_UBSAN, AFL_USE_LSAN:\n"
         "  used in the instrumentation summary message\n",
         argv[0]);
 
@@ -558,6 +597,10 @@ int main(int argc, char** argv) {
   gettimeofday(&tv, &tz);
 
   rand_seed = tv.tv_sec ^ tv.tv_usec ^ getpid();
+  // in fast systems where pids can repeat in the same seconds we need this
+  for (i = 1; (s32)i < argc; i++)
+    for (j = 0; j < strlen(argv[i]); j++)
+      rand_seed += argv[i][j];
 
   srandom(rand_seed);
 
@@ -565,14 +608,20 @@ int main(int argc, char** argv) {
 
   if (inst_ratio_str) {
 
-    if (sscanf(inst_ratio_str, "%u", &inst_ratio) != 1 || inst_ratio > 100)
+    if (sscanf(inst_ratio_str, "%u", &inst_ratio) != 1 || inst_ratio > 100) {
+
       FATAL("Bad value of AFL_INST_RATIO (must be between 0 and 100)");
 
+    }
+
   }
 
-  if (getenv(AS_LOOP_ENV_VAR))
+  if (getenv(AS_LOOP_ENV_VAR)) {
+
     FATAL("Endless loop when calling 'as' (remove '.' from your PATH)");
 
+  }
+
   setenv(AS_LOOP_ENV_VAR, "1", 1);
 
   /* When compiling with ASAN, we don't have a particularly elegant way to skip
@@ -582,24 +631,24 @@ int main(int argc, char** argv) {
   if (getenv("AFL_USE_ASAN") || getenv("AFL_USE_MSAN")) {
 
     sanitizer = 1;
-    if (!getenv("AFL_INST_RATIO")) inst_ratio /= 3;
+    if (!getenv("AFL_INST_RATIO")) { inst_ratio /= 3; }
 
   }
 
-  if (!just_version) add_instrumentation();
+  if (!just_version) { add_instrumentation(); }
 
   if (!(pid = fork())) {
 
-    execvp(as_params[0], (char**)as_params);
+    execvp(as_params[0], (char **)as_params);
     FATAL("Oops, failed to execute '%s' - check your PATH", as_params[0]);
 
   }
 
-  if (pid < 0) PFATAL("fork() failed");
+  if (pid < 0) { PFATAL("fork() failed"); }
 
-  if (waitpid(pid, &status, 0) <= 0) PFATAL("waitpid() failed");
+  if (waitpid(pid, &status, 0) <= 0) { PFATAL("waitpid() failed"); }
 
-  if (!getenv("AFL_KEEP_ASSEMBLY")) unlink(modified_file);
+  if (!getenv("AFL_KEEP_ASSEMBLY")) { unlink(modified_file); }
 
   exit(WEXITSTATUS(status));
 
